"""
"""


#[
from __future__ import annotations

from typing import (Self, Any, )
import prettytable as _pt
#]


class PrettyMixin:
    @property
    def pretty(self, /, ) -> _pt.PrettyTable:
        """
        """
        return self.get_pretty()

    @property
    def pretty_full(self, /, ) -> str:
        """
        """
        return self.get_pretty(full=True, )

    def get_pretty(
        self,
        /,
        full: bool = False,
    ) -> _pt.PrettyTable:
        """
        """
        table = _pt.PrettyTable()
        table.field_names = ("", ) + tuple("{:>10}".format(table) for table in self.base_range)
        table.align = "r"
        table.align[""] = "l"
        for r in ("exogenized", "endogenized", "anticipated", ):
            if getattr(self, f"_{r}_register"):
                _add_register_to_table(
                    table,
                    getattr(self, f"_{r}_register"),
                    getattr(self, f"_default_{r}"),
                    full,
                )
        return table

    def pretty_print(self, *args, **kwargs, ) -> None:
        """
        """
        print(self.get_pretty_table(*args, **kwargs, ), )

    def get_pretty_string(self, *args, **kwargs, ) -> str:
        """
        """
        return self.get_pretty_table(*args, **kwargs, ).get_string()


def _add_register_to_table(
    table,
    register,
    default,
    full,
) -> None:
    """
    """
    #[
    previous = None
    for k, v in register.items():
        if full or any(v):
            points = [ _PRETTY_SYMBOL.get(i, _PRETTY_SYMBOL[Any]) for i in v ]
            if previous:
                table.add_row(previous, )
            previous = [k] + points
    if previous:
        table.add_row(previous, divider=True, )
    #]


_PRETTY_SYMBOL = {
    None: "",
    True: "T",
    False: "F",
    Any: "◼︎",
}




"""
"""


#[
from __future__ import annotations

from numbers import (Number, )
import copy as _co

from .conveniences import iterators as _iterators
#]


class IterVariantsMixin:
    """
    Mixin for handling multiple alternative variants of a single model object
    """
    #[

    @property
    def num_variants(self, /, ) -> int:
        """
        Number of alternative variants within this model
        """
        return len(self._variants)

    @property
    def is_singleton(self, /, ) -> bool:
        """
        True for models with only one variant
        """
        return self.num_variants == 1

    def new_with_shallow_variants(self, /, ) -> Self:
        """
        Create a new model with pointers to invariant and variants of this model
        """
        new = type(self)()
        new._invariant = self._invariant
        new._variants = [ v for v in self._variants ]
        return new

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> Self:
        """
        Alter (expand, shrink) the number of alternative parameter variants in this model object
        """
        if new_num < self.num_variants:
            self.shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self.expand_num_variants(new_num, )
        return self

    def shrink_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        if new_num<1:
            raise Exception('Number of variants must be one or more')
        self._variants = (
            self._variants[0:new_num]
            if new_num < self.num_variants
            else self._variants
        )

    def expand_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        for i in range(self.num_variants, new_num):
            self._variants.append(_co.deepcopy(self._variants[-1]))

    def get_variant(
        self,
        vids: Iterable[int] | int | slice | EllipsisType,
        /,
    ) -> Self:
        new = self.new_with_shallow_variants()
        variant_iter = _resolve_vids(new, vids, )
        new._variants = [ new._variants[i] for i in variant_iter ]
        return new

    def iter_variants(
        self,
        /,
    ) -> Iterator[Self]:
        """
        """
        yield from _iterators.exhaust_then_last(self, )

    def _new_with_single_variant(
        self,
        variant: _variants.Variant,
        /,
    ) -> Self:
        new = self.new_with_shallow_variants()
        new._variants = [ variant ]
        return new

    def __iter__(
        self,
        /,
    ) -> Iterator[Self]:
        """
        Iterate over alternative variants of this model object
        """
        for v in self._variants:
            yield self._new_with_single_variant(v, )

    #]


def _resolve_vids(
    self,
    vids: Iterable[int] | int | slice | EllipsisType,
    /,
) -> Iterable[int]:
    """
    """
    #[
    if isinstance(vids, Number):
        return (int(vids), )
    elif vids is ...:
        return range(self.num_variants, )
    elif isinstance(vids, slice):
        return range(*vids.indices(self.num_variants, ), )
    else:
        return tuple(int(i) for i in vids)
    #]




"""
"""


#[
from __future__ import annotations

from typing import (Protocol, Callable, )
from numbers import (Number, )
import numpy as _np

from ..incidences import main as _incidence
#]


_WHEN_DATA_SYMBOL = {
    True: "?",
    False: "!",
    None: "",
}


class PlanTransformProtocol(Protocol, ):
    """
    """
    #[

    _DATABOX_NAME: str
    _SYMBOL: str

    def eval_exogenized(
        self,
        exogenized_value: Number,
        lagged: Number,
    ) -> Number:
        ...

    def __str__(self, ) -> str: ...

    #]


class PlanTransform:
    """
    """
    #[

    def __init__(
        self,
        /,
        when_data: bool | None = False,
    ) -> None:
        self.when_data = when_data

    def __str__(self, /, ) -> str:
        return _WHEN_DATA_SYMBOL[self.when_data] + self._SYMBOL

    def __repr__(self, /, ) -> self:
        return self.__str__()

    def resolve_databox_name(
        self,
        base_name: str, 
        /,
    ) -> str | None:
        """
        """
        return (
            self._DATABOX_NAME.format(base_name, )
            if self._DATABOX_NAME is not None else None
        )

    #]


class PlanTransformNone(PlanTransform, ):
    """
    """
    #[
    _DATABOX_NAME = "{}"
    _SYMBOL = "[•]"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        values_before: _np.ndarray,
        /,
    ) -> Number:
        return exogenized_value
    #]


class PlanTransformLog(PlanTransform, ):
    """
    """
    #[
    _DATABOX_NAME = "log_{}"
    _SYMBOL = "[log]"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        values_before: _np.ndarray,
        /,
    ) -> Number:
        return _np.exp(exogenized_value)
    #]


class PlanTransformDiff(PlanTransform, ):
    """
    """
    #[
    _DATABOX_NAME = "diff_{}"
    _SYMBOL = "[diff]"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        values_before: _np.ndarray,
        /,
    ) -> Number:
        return values_before[-1] + exogenized_value
    #]


class PlanTransformDiffLog(PlanTransform, ):
    """
    """
    #[
    _DATABOX_NAME = "diff_log_{}"
    _SYMBOL = "[diff_log]"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        values_before: _np.ndarray,
        /,
    ) -> Number:
        return values_before[-1] * _np.exp(exogenized_value)
    #]


class PlanTransformRoc(PlanTransform, ):
    """
    """
    #[
    _DATABOX_NAME = "roc_{}"
    _SYMBOL = "[roc]"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        values_before: _np.ndarray,
        /,
    ) -> Number:
        return values_before[-1] * exogenized_value
    #]


class PlanTransformPct(PlanTransform, ):
    """
    """
    #[

    _DATABOX_NAME = "pct_{}"
    _SYMBOL = "[pct]"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        values_before: _np.ndarray,
        /,
    ) -> Number:
        return values_before[-1] * (1 + exogenized_value/100)

    #]


class PlanTransformFlat(PlanTransform, ):
    """
    """
    #[

    _DATABOX_NAME = None
    _SYMBOL = "[flat]"

    def eval_exogenized(
        self,
        exogenized_value: None,
        values_before: _np.ndarray,
        /,
    ) -> Number:
        return values_before[-1]

    #]


class PlanTransformFactory(PlanTransformNone, ):
    """
    """
    #[

    def __init__(
        self,
        eval_exogenized: Callable,
        databox_name: str,
        /,
        symbol: str = "[custom]",
        when_data: bool | None = False,
    ) -> None:
        """
        """
        self.eval_exogenized = eval_exogenized
        self._DATABOX_NAME = databox_name
        self._SYMBOL = symbol
        self.when_data = when_data
    #]


_TRANFORM_CLASS_FACTORY = {
    None: PlanTransformNone,
    "level": PlanTransformNone,
    "none": PlanTransformNone,
    "log": PlanTransformLog,
    "diff": PlanTransformDiff,
    "diff_log": PlanTransformDiffLog,
    "difflog": PlanTransformDiffLog,
    "roc": PlanTransformRoc,
    "pct": PlanTransformPct,
    "flat": PlanTransformFlat,
}


def resolve_transform(transform, ) -> PlanTransformProtocol:
    """
    """
    #[
    if transform is None or isinstance(transform, str):
        return _TRANFORM_CLASS_FACTORY[transform]()
    else:
        return transform
    #]




"""
Exporting data to CSV sheets
"""


#[
from __future__ import annotations

import pickle as _pk

from collections.abc import (Iterable, )
from typing import (Any, )
import csv as _cs
import numpy as _np
import itertools as _it
import dataclasses as _dc
import functools as _ft

from .. import dates as _dates
from .. import wrongdoings as _wrongdoings
from ..databoxes import main as _databoxes
#]


_DEFAULT_ROUND = 12


@_dc.dataclass
class _ExportBlock:
    """
    """
    #[

    databox: _databoxes.Databox | None = None
    frequency: _dates.Frequency | None = None
    dates: tuple[_dates.Dater] | None = None
    names: tuple[str] | None = None
    #
    # Options
    total_num_data_rows: int | None = None
    description_row: bool | None = None
    delimiter: str | None = None
    numeric_format: str | None = None
    nan_str: str | None = None
    round: int | None = None

    def __iter__(self, ):
        """
        """
        def _round(x, /, ):
            if self.round is None:
                return x
            else:
                return _np.round(x, self.round)
        #
        descriptions = _get_descriptions_for_names(self.databox, self.names, )
        num_data_columns = _get_num_data_columns_for_names(self.databox, self.names, )
        data_array = _get_data_array_for_names(self.databox, self.names, self.dates, )
        #
        # Empty row
        empty_row = ("", ) + ("", )*sum(num_data_columns) + ("", )
        #
        # Name row
        name_row = tuple(_it.chain.from_iterable( 
            (n, ) + ("*", )*(num_data_columns[i] - 1)
            for i, n in enumerate(self.names, )
        ))
        yield (_get_frequency_mark(self.frequency), ) + name_row + ("", )
        #
        # Description row
        if self.description_row:
            description_row = tuple(_it.chain.from_iterable( 
                (n, ) + ("*", )*(num_data_columns[i] - 1)
                for i, n in enumerate(descriptions, )
            ))
            yield ("", ) + description_row + ("", )
        #
        # Dates and data, row by row
        for date, data_row in zip(self.dates, data_array, ):
            yield \
                (str(date), ) \
                + tuple(x if not _np.isnan(x) else self.nan_str for x in _round(data_row).tolist()) \
                + ("", )
        #
        # Empty rows afterwards
        for _ in range(self.total_num_data_rows - len(self.dates), ):
            yield empty_row

    #]


class ExportMixin:
    """
    Databox mixin for exporting data to disk files
    """
    #[
    def to_sheet(
        self,
        file_name: str,
        *,
        frequency_range: dict | None = None,
        names: Iterable[str] | None = None,
        description_row: bool = False,
        frequency: _dates.Frequency | None = None,
        delimiter: str = ",",
        numeric_format: str = "g",
        nan_str: str = "",
        round: int = _DEFAULT_ROUND,
        csv_writer_settings: dict = None,
        when_empty: Literal["error", "warning", "silent"] = "warning",
    ) -> dict[str, Any]:
        """
        """
        frequency_range = _resolve_frequency_range(self, frequency_range, )
        frequency_names = _resolve_frequency_names(self, frequency_range, names, )
        _catch_empty(frequency_range, frequency_names, when_empty, file_name, )
        #
        total_num_data_rows = _get_total_num_data_rows(frequency_range, )
        export_blocks = (
            _ExportBlock(
                self, frequency, frequency_range[frequency], frequency_names[frequency],
                total_num_data_rows, description_row, delimiter, numeric_format, nan_str, round,
            )
            for frequency in frequency_range.keys()
        )
        #
        csv_writer_settings = {} if csv_writer_settings is None else csv_writer_settings
        with open(file_name, "w+") as fid:
            writer = _cs.writer(fid, delimiter=delimiter, **csv_writer_settings)
            for row in zip(*export_blocks, ):
                writer.writerow(_it.chain.from_iterable(row))
        #
        info = {
            "names_exported": tuple(_it.chain.from_iterable(frequency_names.values())),
        }
        #
        return info

    def to_pickle(
        self,
        file_name: str,
        /,
        **kwargs,
    ) -> None:
        """
        """
        with open(file_name, "wb+") as fid:
            _pk.dump(self, fid, **kwargs, )
    #]


def _get_frequency_mark(frequency, ):
    return "__" + frequency.name.lower() + "__"


def _get_names_to_export(databox, frequency, names, ):
    """
    """
    #[
    names_from_databox = self.get_series_names_by_frequency(frequency)
    names = names_from_databox if names is None else names
    return [n for n in names if n in names_from_databox]
    #]


_DEFAULT_FREQUENCY_RANGE = {
    _dates.Frequency.YEARLY: ...,
    _dates.Frequency.HALFYEARLY: ...,
    _dates.Frequency.QUARTERLY: ...,
    _dates.Frequency.MONTHLY: ...,
    _dates.Frequency.DAILY: ...,
    _dates.Frequency.INTEGER: ...,
    _dates.Frequency.UNKNOWN: ...,
}


def _resolve_frequency_range(
    self,
    frequency_range: dict[_dates.Frequency | int: Iterable[_dates.Dater]] | None,
    /,
) -> tuple[dict[_dates.Frequency: tuple[_dates.Dater]], int]:
    """
    """
    #[
    # Remove Nones, ensure Frequency objects
    frequency_range = {
        _dates.Frequency(k): v
        for k, v in (frequency_range or _DEFAULT_FREQUENCY_RANGE).items()
        if v is not None
    }
    #
    # Resolve date range for each ...
    frequency_range = {
        k: v if v is not ... else self.get_range_by_frequency(k, )
        for k, v in frequency_range.items()
    }
    #
    # Expand date ranges into tuples, remove empty ranges
    frequency_range = {
        k: tuple(i for i in v)
        for k, v in frequency_range.items()
        if v is not _dates.EmptyRanger() or k is _dates.Frequency.UNKNOWN
    }
    #
    return frequency_range
    #]


def _resolve_frequency_names(
    self,
    frequency_range: _dates.Frequency | None,
    custom_names: Iterable[str] | None,
    /,
) -> dict[_dates.Frequency: tuple[str, ...]]:
    """
    """
    #[
    #
    # Get all databox names for each frequency
    frequency_names = {
        k: tuple(self.get_series_names_by_frequency(k, ))
        for k in frequency_range.keys()
    }
    #
    # Filter names against custom names
    frequency_names = {
        k: tuple(n for n in v if n in custom_names)
        for k, v in frequency_names.items()
    } if custom_names is not None else frequency_names
    #
    return frequency_names


def _get_total_num_data_rows(
    frequency_range: dict[_dates.Frequency: tuple[_dates.Dater]],
    /,
) -> int:
    """
    """
    #
    # Find maximum number of data rows
    return max((len(v) for v in frequency_range.values()), default=0, )


def _get_descriptions_for_names(
    self,
    names: Iterable[str],
    /,
) -> tuple[str, ...]:
    """
    """
    return tuple(self[n].get_description() for n in names)


def _get_num_data_columns_for_names(
    self,
    names: Iterable[str],
    /,
) -> tuple[int, ...]:
    """
    """
    return tuple(self[n].shape[1] for n in names)


def _get_data_array_for_names(
    self,
    names: Iterable[str],
    dates: Iterable[_dates.Dater],
    /,
) -> _np.ndarray:
    """
    """
    empty_lead = _np.empty((len(dates), 0, ), dtype=_np.float64, )
    return _np.hstack([empty_lead] + [self[n].get_data(dates, ) for n in names], )


def _catch_empty(
    frequency_range: dict[_dates.Frequency: tuple[_dates.Dater]],
    frequency_names: dict[_dates.Frequency: tuple[str, ...]],
    when_empty: Literal["error", "warning", "silent"],
    file_name: str,
    /,
) -> None:
    """
    """
    #[
    if not frequency_range or all(len(v) == 0 for v in frequency_names.values()):
        _wrongdoings.raise_as(when_empty, f"No data exported to {file_name}", )
    #]

